SCHEDULER_QUEUE_KEY = '%(spider)s_requests'
SCHEDULER_QUEUE_CLASS = 'gerapy_rabbitmq.queue.PriorityQueue'
SCHEDULER_QUEUE_MAX_PRIORITY = 100
SCHEDULER_QUEUE_PRIORITY_OFFSET = 30
SCHEDULER_QUEUE_DURABLE = True
SCHEDULER_QUEUE_FORCE_FLUSH = False
SCHEDULER_PERSIST = False
SCHEDULER_IDLE_BEFORE_CLOSE = 0
SCHEDULER_FLUSH_ON_START = False
SCHEDULER_PRE_ENQUEUE_ALL_START_REQUESTS = True

DUPEFILTER_CLASS = 'gerapy_redis.dupefilter.RFPDupeFilter'

RABBITMQ_CONNECTION_PARAMETERS = {'host': 'localhost'}
RABBITMQ_URL_PARAMETERS = ''